﻿using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON;
using MacGruber;

namespace bill_prime
{
    public class ClothingTriggers : MVRScript
    {
        private JSONStorableString mySelectedClothingItem;
        private JSONStorableStringChooser myClothingTriggersChooser;
        private JSONStorableBool myAcceptTrigger;
        private JSONStorableString myChoiceName;
        private JSONStorableString mySelectClothingItem;
        private DAZCharacterSelector myCharacter;
        private JSONStorableStringChooser myClothingList;
        private UIDynamicPopup myClothingListPopup;
        private List<object> myClothingTriggersUI = new List<object>();
        private Dictionary<string, string> myUidLookup = new Dictionary<string, string>();
        private float myLoadingClock = 0.1f;

        private class ClothingOnOffTrigger
        {
            public ClothingOnOffTrigger(MVRScript script, string name, string uid)
            {
                ClothingName = name;
                var geometry = script.containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;

                if (uid.Contains(".latest:"))
                {
                    myClothingItem = geometry.clothingItems.FirstOrDefault(c => c.displayName == name && c.isLatestVersion && c.uid == uid.Replace(".latest:", "." + c.version.Replace("v", "") + ":"));
                }
                else
                {
                    myClothingItem = geometry.clothingItems.FirstOrDefault(c => c.uid == uid);
                }

                ClothingUid = myClothingItem.uid;
                myPreviousState = myClothingItem.isActiveAndEnabled;
                myClothSimControls = myClothingItem.GetComponentsInChildren<ClothSimControl>();
                myPreviousUndressState = myClothSimControls.Any(o => o.clothSettings.BreakEnabled);
                TriggerWear = new EventTrigger(script, "OnWear");
                TriggerRemove = new EventTrigger(script, "OnRemove");
                myWearTriggerWaiting = false;
                myRemoveTriggerWaiting = false;
            }

            public ClothingOnOffTrigger(MVRScript script, string name, DAZClothingItem clothingItem)
            {
                ClothingName = name;
                myClothingItem = clothingItem;
                ClothingUid = clothingItem.uid;
                myPreviousState = clothingItem.isActiveAndEnabled;
                myClothSimControls = clothingItem.GetComponentsInChildren<ClothSimControl>();
                myPreviousUndressState = myClothSimControls.Any(o => o.clothSettings.BreakEnabled);
                TriggerWear = new EventTrigger(script, "OnWear");
                TriggerRemove = new EventTrigger(script, "OnRemove");
                myWearTriggerWaiting = false;
                myRemoveTriggerWaiting = false;
                myClockStarted = false;
            }

            public ClothingOnOffTrigger(MVRScript script, string name, DAZClothingItem clothingItem, ClothingOnOffTrigger source)
            {
                myClothingItem = clothingItem;
                ClothingUid = clothingItem.uid;
                myPreviousState = clothingItem.isActiveAndEnabled;
                myClothSimControls = clothingItem.GetComponentsInChildren<ClothSimControl>();
                myPreviousUndressState = myClothSimControls.Any(o => o.clothSettings.BreakEnabled);
                TriggerWear = new EventTrigger(source.TriggerWear);
                TriggerRemove = new EventTrigger(source.TriggerRemove);
                Name = name;
                myWearTriggerWaiting = false;
                myRemoveTriggerWaiting = false;
                myClockStarted = false;
            }

            public static int SortByNameAscending(ClothingOnOffTrigger a, ClothingOnOffTrigger b)
            {
                return a.Name.CompareTo(b.Name);
            }

            public string Name
            {
                get { return ClothingName; }
                set
                {
                    ClothingName = value;
                    TriggerWear.Name = "OnWear";
                    TriggerRemove.Name = "OnRemove";
                }
            }

            public void EvaluateStateChange(float loadingClock)
            {
                if (ClothingName == null)
                    return;

                if (loadingClock <= 0.0)
                {
                    //clothing item state has changed
                    myClothSimControls = myClothingItem.GetComponentsInChildren<ClothSimControl>();
                    bool anyUndressed = myClothSimControls.Any(o => o.clothSettings.BreakEnabled);

                    if (myClothingItem.isActiveAndEnabled != myPreviousState
                    || anyUndressed != myPreviousUndressState)
                    {
                        if (!myClockStarted)
                        {
                            myClockStarted = true;
                            myToggleClock = 0.01f;
                        }
                        else if (myClockStarted && myToggleClock > 0.0)
                        {
                            myToggleClock -= Time.deltaTime;
                        }
                        else
                        {
                            myClockStarted = false;
                            myPreviousState = myClothingItem.isActiveAndEnabled;
                            myPreviousUndressState = anyUndressed;

                            if (myClothingItem.isActiveAndEnabled && !anyUndressed)
                                myWearTriggerWaiting = true;
                            else
                                myRemoveTriggerWaiting = true;

                            //OnStateChanged?.Invoke(myClothingItem);
                        }
                    }
                }
            }

            public EventTrigger TriggerWear;
            public EventTrigger TriggerRemove;
            public string ClothingName;
            public string ClothingUid;
            private bool myPreviousState;
            private bool myPreviousUndressState;
            public bool myWearTriggerWaiting;
            public bool myRemoveTriggerWaiting;
            private bool myClockStarted;
            private float myToggleClock;
            private DAZClothingItem myClothingItem { get; }
            private ClothSimControl[] myClothSimControls;
            public delegate void OnStateChangedEvent(DAZClothingItem clothingItem);
            public OnStateChangedEvent OnStateChanged;
        }

        private List<ClothingOnOffTrigger> myClothingTriggers = new List<ClothingOnOffTrigger>();
        private int myClothingItemIndex = -1;
        private bool myNeedUIRefresh = false;

        public override void Init()
        {
            if (containingAtom.type != "Person")
            {
                SuperController.LogMessage("ClothingTriggers (Disabled): This plugin is designed to be used on a person atom.");
                enabled = false;
                return;
            }

            Utils.OnInitUI(CreateUIElement);

            Utils.SetupInfoText(this,
                "<color=#606060><size=40><b>ClothingTriggers</b></size>\nCan trigger defined actions when clothing is worn (<i>OnWear</i>) or Removed/undressed (<i>OnRemove</i>).</color>",
                320.0f, true
            );

            //Get geometry values
            var geometry = containingAtom.GetStorableByID("geometry");
            myCharacter = geometry as DAZCharacterSelector;

            //refresh active clothing in dropdown
            Utils.SetupButton(this, "Refresh Active Clothing", RefreshClothingList, false);

            //create initial clothing list and ui
            var activeClothingNames = myCharacter.clothingItems.Where(c => c.isActiveAndEnabled).Select(c => c.displayName).ToList();
            myClothingList = new JSONStorableStringChooser("ClothingListPopup", activeClothingNames, "", "Active Clothing");
            RegisterStringChooser(myClothingList);
            myClothingListPopup = CreatePopup(myClothingList, false);
            myUidLookup = new Dictionary<string, string>();

            foreach (DAZClothingItem clothing in myCharacter.clothingItems.Where(c => c.isActiveAndEnabled))
            {
                myUidLookup.Add(clothing.displayName, clothing.uid);
            }

            //mySelectedClothingItem = new JSONStorableString("CurrentChoice", "Current Choice:");
            //Utils.SetupInfoOneLine(this, mySelectedClothingItem, true);

            myClothingTriggersChooser = new JSONStorableStringChooser("ClothingTriggersPopop", new List<string>(), "", "Clothing Triggers");
            myClothingTriggersChooser.isStorable = myClothingTriggersChooser.isRestorable = false;
            myClothingTriggersChooser.setCallbackFunction += UISetChoice;
            CreatePopup(myClothingTriggersChooser, false);

            //Utils.SetupTwinButton(this, "Add Trigger", UIAddChoice, "Duplicate Choice", UIDuplicateChoice, false);
            Utils.SetupButton(this, "Add Triggers", UIAddTriggers, false);
            Utils.SetupButton(this, "Remove Triggers", UIRemoveTriggers, false);

            myAcceptTrigger = Utils.SetupToggle(this, "TriggersActive", true, false);

            Utils.SetupButton(this, "Test Trigger OnWear", TestTriggerActionWear, false);
            Utils.SetupButton(this, "Test Trigger OnRemove", TestTriggerActionRemove, false);

            //mySelectClothingItem = new JSONStorableString("SelectChoice", "", SelectChoiceAction);
            //mySelectClothingItem.isStorable = mySelectClothingItem.isRestorable = false;
            //RegisterString(mySelectClothingItem);
            //Utils.SetupAction(this, "TriggerWear", TriggerActionWear);
            //Utils.SetupAction(this, "TriggerRemove", TriggerActionRemove);

            Utils.SetupSpacer(this, 15, false);

            //myChoiceName = new JSONStorableString("Name", "", UIRenameChoice);

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            SimpleTriggerHandler.LoadAssets();
        }

        //private void SelectChoiceAction(string newChoice)
        //{
        //    mySelectClothingItem.valNoCallback = string.Empty;
        //    int selectedIndex = myClothingTriggersChooser.choices.IndexOf(newChoice);
        //    if (selectedIndex >= 0)
        //    {
        //        myClothingItemIndex = selectedIndex;
        //        mySelectedClothingItem.val = "Current Choice: " + newChoice;
        //    }
        //    else
        //    {
        //        mySelectedClothingItem.val = "Current Choice:";
        //        SuperController.LogError($"LogicBricks.SelectChoice: Unknown choice '{newChoice}'.");
        //    }
        //}

        private void UISetChoice(string v)
        {
            Utils.RemoveUIElements(this, myClothingTriggersUI);

            int idx = myClothingTriggersChooser.choices.IndexOf(v);
            if (idx < 0)
                return;

            ClothingOnOffTrigger choice = myClothingTriggers[idx];

            UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Clothing Triggers</b></size>", false);
            myClothingTriggersUI.Add(optionHeader);

            //myChoiceName.valNoCallback = choice.Name;
            //UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myChoiceName, false);
            //myClothingTriggersUI.Add(nameInput);

            UIDynamicButton actionsButtonOn = Utils.SetupButton(this, "Actions " + choice.TriggerWear.Name, choice.TriggerWear.OpenPanel, false);
            myClothingTriggersUI.Add(actionsButtonOn);

            UIDynamicButton actionsButtonOff = Utils.SetupButton(this, "Actions " + choice.TriggerRemove.Name, choice.TriggerRemove.OpenPanel, false);
            myClothingTriggersUI.Add(actionsButtonOff);
        }

        private void UIAddTriggers()
        {
            UIAddTrigger(null);
        }

        private void UIAddTrigger(ClothingOnOffTrigger source)
        {
            if (!myClothingTriggersChooser.choices.Any(c => c == myClothingListPopup.popup.currentValue))
            {
                var clothing = myCharacter.clothingItems.FirstOrDefault(c => c.uid == myUidLookup[myClothingListPopup.popup.currentValue]);
                if (clothing != null)
                {
                    ClothingOnOffTrigger cTriggers;
                    if (source == null)
                        cTriggers = new ClothingOnOffTrigger(this, clothing.displayName, clothing);
                    else
                        cTriggers = new ClothingOnOffTrigger(this, clothing.displayName, clothing, source);
                    cTriggers.OnStateChanged += new ClothingOnOffTrigger.OnStateChangedEvent(OnClothingItemStateChanged);
                    myClothingTriggers.Add(cTriggers);
                    myClothingTriggers.Sort(ClothingOnOffTrigger.SortByNameAscending);
                    List<string> options = myClothingTriggersChooser.choices;
                    options.Clear();
                    for (int e = 0; e < myClothingTriggers.Count; ++e)
                        options.Add(myClothingTriggers[e].Name);
                    myClothingTriggersChooser.choices = null; // force UI sync
                    myClothingTriggersChooser.choices = options;
                    myClothingTriggersChooser.val = clothing.displayName;
                    return;
                }
            }
            SuperController.LogError("ClothingTriggers: Error adding triggers!");
        }

        private void UIDuplicateChoice()
        {
            int idx = myClothingTriggersChooser.choices.IndexOf(myClothingTriggersChooser.val);
            ClothingOnOffTrigger choice = idx >= 0 ? myClothingTriggers[idx] : null;
            UIAddTrigger(choice);
        }

        private void UIRemoveTriggers()
        {
            int idx = myClothingTriggersChooser.choices.IndexOf(myClothingTriggersChooser.val);
            if (idx < 0)
                return;

            List<string> options = myClothingTriggersChooser.choices;
            options.RemoveAt(idx);
            myClothingTriggers[idx].TriggerWear.Remove();
            myClothingTriggers[idx].TriggerRemove.Remove();
            myClothingTriggers.RemoveAt(idx);
            myClothingTriggersChooser.choices = null; // force UI sync
            myClothingTriggersChooser.choices = options;

            if (idx < options.Count)
                myClothingTriggersChooser.val = options[idx];
            else if (options.Count > 0)
                myClothingTriggersChooser.val = options[options.Count - 1];
            else
                myClothingTriggersChooser.val = "";
        }

        //private void UIRenameChoice(string name)
        //{
        //    int idx = myClothingTriggersChooser.choices.IndexOf(myClothingTriggersChooser.val);
        //    if (idx < 0)
        //        return;

        //    myClothingTriggers[idx].Name = name;

        //    myClothingTriggers.Sort(ClothingTriggers.SortByNameAscending);
        //    List<string> options = myClothingTriggersChooser.choices;
        //    options.Clear();
        //    for (int e = 0; e < myClothingTriggers.Count; ++e)
        //        options.Add(myClothingTriggers[e].Name);
        //    myClothingTriggersChooser.choices = null; // force UI sync
        //    myClothingTriggersChooser.choices = options;
        //    myClothingTriggersChooser.val = name;
        //}

        private void OnAtomRename(string oldid, string newid)
        {
            for (int i = 0; i < myClothingTriggers.Count; ++i)
            {
                myClothingTriggers[i].TriggerWear.SyncAtomNames();
                myClothingTriggers[i].TriggerRemove.SyncAtomNames();
            }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;

                JSONArray clist = new JSONArray();
                for (int i = 0; i < myClothingTriggers.Count; ++i)
                {
                    ClothingOnOffTrigger choice = myClothingTriggers[i];
                    JSONClass cc = new JSONClass();
                    cc["Name"] = choice.Name;
                    cc["ClothingUid"] = choice.ClothingUid;
                    cc[choice.TriggerWear.Name] = choice.TriggerWear.GetJSON(base.subScenePrefix);
                    cc[choice.TriggerRemove.Name] = choice.TriggerRemove.GetJSON(base.subScenePrefix);
                    clist.Add("", cc);
                }
                jc["Choices"] = clist;

                //if (myClothingItemIndex >= 0 && myClothingItemIndex < myClothingTriggers.Count)
                //    jc["Selected"] = myClothingTriggers[myClothingItemIndex].Name;
            }

            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical)
            {
                for (int i = 0; i < myClothingTriggers.Count; ++i)
                {
                    myClothingTriggers[i].TriggerWear.Remove();
                    myClothingTriggers[i].TriggerRemove.Remove();
                }
                myClothingTriggers.Clear();

                JSONArray clist = jc["Choices"].AsArray;
                for (int i = 0; i < clist.Count; ++i)
                {
                    JSONClass cc = clist[i].AsObject;
                    ClothingOnOffTrigger choice = new ClothingOnOffTrigger(this, cc["Name"], cc["ClothingUid"]);
                    choice.OnStateChanged += new ClothingOnOffTrigger.OnStateChangedEvent(OnClothingItemStateChanged);
                    choice.TriggerWear.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                    choice.TriggerRemove.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                    myClothingTriggers.Add(choice);
                }

                myClothingTriggers.Sort(ClothingOnOffTrigger.SortByNameAscending);
                List<string> options = myClothingTriggersChooser.choices;
                options.Clear();
                for (int e = 0; e < myClothingTriggers.Count; ++e)
                    options.Add(myClothingTriggers[e].Name);
                myClothingTriggersChooser.choices = null; // force UI sync
                myClothingTriggersChooser.choices = options;

                if (options.Count == 0)
                    myClothingTriggersChooser.valNoCallback = "";
                else if (!options.Contains(myClothingTriggersChooser.val))
                    myClothingTriggersChooser.valNoCallback = options[0];

                myClothingItemIndex = -1;
                //if (jc.HasKey("Selected"))
                //    SelectChoiceAction(jc["Selected"]);

                myNeedUIRefresh = true;
            }
        }

        private void Update()
        {
            for (int i = 0; i < myClothingTriggers.Count; ++i)
            {
                myClothingTriggers[i].TriggerWear.Update();
                myClothingTriggers[i].TriggerRemove.Update();
            }

            if (myNeedUIRefresh)
            {
                myNeedUIRefresh = false;
                myClothingTriggersChooser.setCallbackFunction(myClothingTriggersChooser.val);
            }

            SuperController sc = SuperController.singleton;
            GameObject loadingIcon = sc.loadingIcon.gameObject;
            GameObject loadingUI = sc.loadingUI.gameObject;
            if (sc.freezeAnimation || loadingIcon.activeSelf || loadingUI.activeSelf)
            {
                myLoadingClock = 0.01f;
                return;
            }

            CheckForClothingChanges();

            if (myClothingTriggers.Any(c => c.myRemoveTriggerWaiting) && myLoadingClock <= 0.0)
                ExecuteRemoveClothingTriggers();

            if (myClothingTriggers.Any(c => c.myWearTriggerWaiting) && myLoadingClock <= 0.0)
                ExecuteWearClothingTriggers();

            if (myLoadingClock > 0.0)
                myLoadingClock -= Time.deltaTime;
        }
        private void OnDisable()
        {
            myLoadingClock = 0.0f;
        }

        private void OnDestroy()
        {
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

            for (int i = 0; i < myClothingTriggers.Count; ++i)
            {
                myClothingTriggers[i].TriggerWear.Remove();
                myClothingTriggers[i].TriggerRemove.Remove();
            }

            myClothingTriggers.Clear();

            Utils.OnDestroyUI();
        }

        private void TriggerActionWear(string ClothingName)
        {
            if (myAcceptTrigger.val && myClothingTriggers.Any(c => c.ClothingName == ClothingName))
            {
                myClothingTriggers.First(c => c.ClothingName == ClothingName).TriggerWear.Trigger();
            }
        }


        private void TriggerActionRemove(string ClothingName)
        {
            if (myAcceptTrigger.val && myClothingTriggers.Any(c => c.ClothingName == ClothingName))
            {
                myClothingTriggers.First(c => c.ClothingName == ClothingName).TriggerRemove.Trigger();
            }
        }

        private void TestTriggerActionWear()
        {
            if (myClothingTriggersChooser.val != null && myClothingTriggersChooser.val != "")
            {
                TriggerActionWear(myClothingTriggersChooser.val);
            }
        }

        private void TestTriggerActionRemove()
        {
            if (myClothingTriggersChooser.val != null && myClothingTriggersChooser.val != "")
            {
                TriggerActionRemove(myClothingTriggersChooser.val);
            }
        }

        private void RefreshClothingList()
        {
            myClothingList.choices.Clear();
            myClothingList.choices = myCharacter.clothingItems.Where(c => c.isActiveAndEnabled).Select(c => c.displayName).ToList();
            myClothingList.SetValToDefault();

            myUidLookup = new Dictionary<string, string>();

            foreach (DAZClothingItem clothing in myCharacter.clothingItems.Where(c => c.isActiveAndEnabled))
            {
                myUidLookup.Add(clothing.displayName, clothing.uid);
            }
        }

        private void CheckForClothingChanges()
        {
            foreach (var c in myClothingTriggers)
                c.EvaluateStateChange(myLoadingClock);
        }

        private void ExecuteRemoveClothingTriggers()
        {
            if (myAcceptTrigger.val)
            {
                foreach (var c in myClothingTriggers.Where(c => c.myRemoveTriggerWaiting))
                {
                    c.TriggerRemove.Trigger();
                    c.myRemoveTriggerWaiting = false;
                }

                myLoadingClock = 0.01f;
            }
        }

        private void ExecuteWearClothingTriggers()
        {
            if (myAcceptTrigger.val)
            {
                foreach (var c in myClothingTriggers.Where(c => c.myWearTriggerWaiting))
                {
                    c.TriggerWear.Trigger();
                    c.myWearTriggerWaiting = false;
                }
            }
        }

        public void OnClothingItemStateChanged(DAZClothingItem clothingItem)
        {
            //item is now active so TriggerOnWear
            if (clothingItem.isActiveAndEnabled)
            {
                var clothSimControlChildren = clothingItem.GetComponentsInChildren<ClothSimControl>();
                bool anyUndressed = clothSimControlChildren.Any(o => o.clothSettings.BreakEnabled);

                if (anyUndressed)
                {
                    TriggerActionRemove(clothingItem.displayName);
                    return;
                }
                TriggerActionWear(clothingItem.displayName);
                return;
            }

            TriggerActionRemove(clothingItem.displayName);
        }
    }
}